"""Reusable types for the Picata project."""

from typing import Any, TypedDict

from django.http import HttpRequest

# Generic arguments and keyword arguments
Args = tuple[Any, ...]
Kwargs = dict[str, Any]


class Context(TypedDict):
    """Base class for context dicts passed all around the system."""

    request: HttpRequest


# Log arguments for structured logging
LogArg = str | int | float | bool



"""Custom validators for Django fields."""

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator


class HREFValidator:
    """Custom validator for href attributes in HTML anchor tags.

    Basically it's `URLValidator` but with 'mailto:', 'tel:' and 'file:' added.
    (Doesn't actually validate email addresses, telephone numbers, or file paths.)
    """

    def __init__(
        self, extra_schemes: list[str] | None = None, url_schemes: list[str] | None = None
    ) -> None:
        """Store schemes we check for and initialise a basic `URLValidator`."""
        self.schemes: list[str] = extra_schemes or ["mailto", "tel", "file", "sms"]
        self.base_validator = URLValidator(schemes=url_schemes)

    def __call__(self, value: str) -> None:
        """Try validating a custom scheme (e.g. "mailto:…") if the `URLValidator` fails."""
        try:
            self.base_validator(value)
        except ValidationError as err:
            if ":" in value:
                if not any(value.startswith(f"{scheme}:") for scheme in self.schemes):
                    raise ValidationError(
                        f"'{value}' is not a valid href.", code="invalid_href"
                    ) from err
            elif not value.startswith(("/", "#")):  # Allow relative links or fragments
                raise ValidationError(
                    f"'{value}' is not a valid href.", code="invalid_href"
                ) from err
        else:
            return



"""Template tags for rendering menus."""  # noqa: INP001

# NB: Django's meta-class shenanigans over-complicate type hinting when QuerySets get involved.
# pyright: reportAttributeAccessIssue=false

from typing import TYPE_CHECKING, TypedDict

from django import template
from wagtail.models import Page, Site

from picata.typing import Context

if TYPE_CHECKING:
    from django.http import HttpRequest

register = template.Library()


class SiteMenuContext(TypedDict):
    """Context returned from `render_site_menu` for `site_menu.html`."""

    menu_pages: list[Page]
    current_section: Page | None


@register.inclusion_tag("picata/tags/site_menu.html", takes_context=True)
def render_site_menu(context: Context) -> SiteMenuContext:
    """Fetch the site root and its child pages for the site menu."""
    current_page: Page | None = context.get("self")
    active_section = None

    request: HttpRequest = context["request"]
    current_site = Site.find_for_request(request)
    if not current_site:
        raise ValueError("No Wagtail Site found for the current request.")

    # Get menu pages as a list of pages starting with the root, followed by its children
    root_page = current_site.root_page.specific

    top_pages = root_page.get_children().in_menu()
    if not request.user.is_authenticated:
        top_pages = top_pages.live()
    top_pages = top_pages.specific()

    menu_pages = [root_page, *top_pages]

    # Find which page in the set of menu pages that the current page is a descendant of
    if current_page == root_page:
        active_section = root_page
    else:
        for page in top_pages:
            if current_page and (current_page == page or current_page.is_descendant_of(page)):
                active_section = page
                break

    return {"menu_pages": menu_pages, "current_section": active_section}



"""Application configuration for the Picata Django app."""

from django.apps import AppConfig


class Config(AppConfig):
    """Configuration class for the Picata Django application."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "picata"

    def ready(self) -> None:
        """Configure Wagtail admin with custom models, and register document transformers."""
        #
        # Register the 'custom article type' model with the Wagtail admin
        from wagtail_modeladmin.options import modeladmin_register

        from picata.models import ArticleTypeAdmin

        modeladmin_register(ArticleTypeAdmin)

        # Add document transformers to the HTMLProcessingMiddleware
        from picata.middleware import HTMLProcessingMiddleware
        from picata.transformers import AnchorInserter, add_heading_ids

        ## Add ids to all headings missing them within html > body > main
        HTMLProcessingMiddleware.add_transformer(add_heading_ids)

        ## Add anchored pillcrows to headings in designated pages
        anchor_inserter = AnchorInserter(
            root=".//article", targets=".//h1 | .//h2 | .//h3 | .//h4 | .//h5 | .//h6"
        )
        HTMLProcessingMiddleware.add_transformer(anchor_inserter)



"""Wagtail "blocks"."""

import pygments
from django.forms import CharField
from django.utils.safestring import mark_safe
from pygments import formatters, lexers
from pygments.util import ClassNotFound
from wagtail.blocks import (
    CharBlock,
    ChoiceBlock,
    IntegerBlock,
    ListBlock,
    RichTextBlock,
    StreamBlock,
    StructBlock,
    TextBlock,
    URLBlock,
)
from wagtail.images.blocks import ImageChooserBlock

from picata.typing.wagtail import BlockRenderContext, BlockRenderValue
from picata.validators import HREFValidator


class HREFField(CharField):
    """Custom field for href attributes (i.e. URLs but also schemes like 'mailto:')."""

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Initialise a CharField with `HREFValidator` as the only validator."""
        kwargs["validators"] = [HREFValidator()]
        super().__init__(*args, **kwargs)


class HREFBlock(URLBlock):
    """Custom block for href attributes (i.e. a `URLBlock` using the extended `HREFField`)."""

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Initialise the `URLBlock` superclass and replace our field with a `HREFField`."""
        super().__init__(*args, **kwargs)
        self.field = HREFField()


class StaticIconLinkItemBlock(StructBlock):
    """A single list item with an optional icon and surrounding anchor."""

    href = HREFBlock(
        required=False,
        help_text="An optional link field",
        max_length=255,
    )
    label = CharBlock(required=True, max_length=50, help_text="The title for the list item.")
    icon = CharBlock(
        required=False,
        max_length=255,
        help_text="Id of the icon in the static/icons.svg file.",
    )

    class Meta:
        """Meta information."""

        template = "picata/blocks/icon_link_item.html"


class StaticIconLinkListBlock(StructBlock):
    """A list of optionally-linked list items with an optional heading."""

    heading = CharBlock(
        required=False,
        help_text="Optional heading for this list (e.g., Social Links).",
    )
    heading_level = IntegerBlock(
        required=False,
        min_value=1,
        max_value=6,
        default=2,
        help_text="Heading level for the list (1-6).",
    )
    items = ListBlock(
        StaticIconLinkItemBlock(),
        help_text="The list of items.",
    )

    class Meta:
        """Meta information."""

        template = "picata/blocks/icon_link_list.html"


class StaticIconLinkListsBlock(StructBlock):
    """A wrapper for multiple heading-and-link-list blocks."""

    lists = StreamBlock(
        [
            ("link_list", StaticIconLinkListBlock()),
        ],
        required=False,
        help_text="Add one or more heading-and-link-list blocks.",
    )

    class Meta:
        """Meta information."""

        template = "picata/blocks/icon_link_list_stream.html"


class CodeBlock(StructBlock):
    """A block for displaying code with optional syntax highlighting."""

    code = TextBlock(required=True, help_text=None)
    language = ChoiceBlock(
        required=False,
        choices=[
            ("plaintext", "Plain Text"),
            ("bash", "Bash"),
            ("css", "CSS"),
            ("javascript", "JavaScript"),
            ("html", "HTML"),
            ("python", "Python"),
            ("toml", "TOML"),
            ("yaml", "YAML"),
        ],
        help_text=None,
    )

    def render_basic(self, value: BlockRenderValue, context: BlockRenderContext = None) -> str:
        """Render the code block with syntax highlighting."""
        code = value.get("code", "")
        language = value.get("language", "plaintext")

        try:
            lexer = lexers.get_lexer_by_name(language)
            formatter = formatters.HtmlFormatter(cssclass="pygments")
            highlighted_code = pygments.highlight(code, lexer, formatter)
        except ClassNotFound:
            highlighted_code = f"<pre><code>{code}</code></pre>"

        return mark_safe(highlighted_code)  # noqa: S308

    class Meta:
        """Meta information."""

        icon = "code"
        label = "Code Block"


class SectionBlock(StructBlock):
    """A page section, with a heading rendered at a defined level."""

    heading = CharBlock(
        required=True, help_text='Heading for this section, included in "page contents".'
    )
    level = IntegerBlock(required=True, min_value=1, max_value=6, help_text="Heading level")
    content = StreamBlock(
        [
            ("rich_text", RichTextBlock()),
            ("image", ImageChooserBlock()),
        ],
        required=False,
        help_text=None,
    )

    class Meta:
        """Meta-info for the block."""

        icon = "folder"
        label = "Section"


class WrappedImageChooserBlock(ImageChooserBlock):
    """An ImageChooserBlock that wraps the output in a div."""

    def render_basic(self, value: BlockRenderValue, context: BlockRenderContext = None) -> str:
        """Render the image wrapped in a div with a custom class."""
        if not value:  # If no image is selected, return an empty string
            return ""

        # Use Wagtail's default rendering for the image, wrapped in a <div>
        image_tag = super().render_basic(value, context)
        return f'<div class="image-wrapper">{image_tag}</div>'



"""Types for Wagtail.

NB: This module split the picata.typing module into a package in order to avoid
circular imports during program initialisation, with the `from wagtail` imports.
"""

from typing import Any

from wagtail.blocks import StructBlock
from wagtail.models import Page

from . import Context


# StructBlock types
class BlockRenderContextDict(Context):
    """Context dicts passed to render functions for Blocks."""

    self: StructBlock
    page: Page


BlockRenderContext = BlockRenderContextDict | None
BlockRenderValue = dict[str, Any]


class PageContext(Context):
    """Base class for Wagtail Page classes."""

    self: Page
    page: Page



#!/usr/bin/env python
"""Entry-point for Django management commands."""

from os import environ
from sys import argv

if __name__ == "__main__":
    environ.setdefault("DJANGO_SETTINGS_MODULE", "picata.settings.dev")

    if len(argv) >= 2:  # noqa: PLR2004
        environ.setdefault("DJANGO_MANAGEMENT_COMMAND", argv[1])

    from django.core.management import execute_from_command_line

    execute_from_command_line(argv)



"""Filter to wrap the last two words of text in a span which prevents word-wrapping."""  # noqa: INP001

from html.parser import HTMLParser

from django.utils.html import escape
from django.utils.safestring import mark_safe


class KillOrphansParser(HTMLParser):
    """Parser that wraps the last two words of visible text in a span tag."""

    result: list[str]
    text_chunks: list[str]

    def __init__(self) -> None:
        """Initialize the parser and result containers."""
        super().__init__()
        self.result = []
        self.text_chunks = []

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        """Handle start tags and store them."""
        start_tag_text = self.get_starttag_text()
        if start_tag_text is not None:
            self.result.append(start_tag_text)

    def handle_endtag(self, tag: str) -> None:
        """Handle end tags and store them."""
        self.result.append(f"</{tag}>")

    def handle_data(self, data: str) -> None:
        """Handle text data and store it."""
        self.text_chunks.append(data)
        self.result.append(data)

    def get_wrapped_html(self) -> str:
        """Return HTML with the last two words wrapped in a span."""
        full_text = "".join(self.text_chunks)
        words = full_text.split()

        if len(words) < 2:  # noqa: PLR2004
            return "".join(self.result)

        wrapped_words = (
            " ".join(words[:-2])
            + f' <span class="whitespace-nowrap">{escape(" ".join(words[-2:]))}</span>'
        )

        return mark_safe(  # noqa: S308
            "".join(wrapped_words if chunk in full_text else chunk for chunk in self.result)
        )


def killorphans(value: str) -> str:
    """Wrap the last two words of visible text in a span tag."""
    parser = KillOrphansParser()
    parser.feed(value)
    return parser.get_wrapped_html()



"""Customisations to the global logging system and custom logging classses."""

import logging
from logging.handlers import TimedRotatingFileHandler

# Define custom levels
TRACE_LEVEL = 5
SUCCESS_LEVEL = 25

logging.addLevelName(TRACE_LEVEL, "TRACE")
logging.addLevelName(SUCCESS_LEVEL, "SUCCESS")


# Handlers


class RotatingDailyFileHandler(TimedRotatingFileHandler):
    """Rotates file logs daily at midnight, with a week of backups by default."""

    def __init__(self, filename: str, backup_days: int = 7) -> None:
        """Pass our defaults and given args to TimedRotatingFileHandler's __init__ method."""
        super().__init__(filename, when="midnight", interval=1, backupCount=backup_days)


# Formatters


class FormatterWithEverything(logging.Formatter):
    """The *most* verbose logging formatter."""

    def format(self, record: logging.LogRecord) -> str:
        """Return a ludicrously verbose log message format."""
        verbose_message = (
            f"LEVEL: {record.levelname}\n"
            f"TIME: {self.formatTime(record)}\n"
            f"LOGGER: {record.name}\n"
            f"MODULE: {record.module}\n"
            f"FUNC: {record.funcName}\n"
            f"LINE: {record.lineno}\n"
            f"PATH: {record.pathname}\n"
            f"THREAD: {record.threadName}\n"
            f"PROCESS: {record.process}\n"
            f"MESSAGE: {record.getMessage()}\n"
        )
        if record.exc_info:
            verbose_message += f"EXCEPTION: {self.formatException(record.exc_info)}\n"
        return verbose_message



"""Generic helper-functions."""
# NB: Django's meta-class shenanigans over-complicate type hinting when QuerySets get involved.
# pyright: reportAttributeAccessIssue=false

import re
from ipaddress import AddressValueError, IPv4Address

from django.apps import apps
from django.db.models import Model
from django.http import HttpResponse, StreamingHttpResponse
from lxml.etree import _Element

# Pre-compile commonly used regular expressions
ALPHANUMERIC_REGEX = re.compile(r"[^a-zA-Z0-9]")


def get_models_of_type(base_type: type[Model]) -> list[type[Model]]:
    """Retrieve all concrete subclasses of the given base Model type."""
    all_models = apps.get_models()

    return [
        model
        for model in all_models
        if issubclass(model, base_type) and not model._meta.abstract  # noqa: SLF001
    ]


def get_public_ip() -> IPv4Address | None:
    """Fetch the public-facing IP of the current host."""
    import socket

    import psutil

    for addrs in psutil.net_if_addrs().values():
        for addr in addrs:
            if addr.family == socket.AF_INET:
                ip = addr.address
                if not ip.startswith(("10.", "192.168.", "172.", "127.")):
                    try:
                        return IPv4Address(ip)
                    except AddressValueError:
                        pass
    return None


def get_full_text(element: _Element) -> str:
    """Extract text from an element and its descendants, concatenate it, and trim whitespace."""
    return "".join(element.xpath(".//text()")).strip()


def make_response(
    original_response: HttpResponse,
    new_content: str | bytes,
) -> HttpResponse:
    """Create a new HttpResponse while preserving attributes from the original response."""
    if isinstance(original_response, StreamingHttpResponse):
        raise TypeError("StreamingHttpResponse objects are not supported.")
    new_response = HttpResponse(
        content=new_content,
        content_type=original_response.get("Content-Type", None),
        status=original_response.status_code,
    )
    for key, value in original_response.headers.items():
        new_response[key] = value
    new_response.cookies = original_response.cookies
    for attr in dir(original_response):
        if not attr.startswith("_") and not hasattr(HttpResponse, attr):
            setattr(new_response, attr, getattr(original_response, attr))

    return new_response



# ruff: noqa: F405 ERA001
# mypy: disable-error-code="index"
# pyright: reportCallIssue=false, reportArgumentType=false
"""Django settings for development environments."""

import logging

from .base import *  # noqa: F403
from .base import LOGGING

# NB: The logging system isn't set up yet; this is the "root" logger, which'll just write to stderr
logger = logging.getLogger()

DEBUG = True
USE_X_FORWARDED_HOST = True

# Security
SECRET_KEY = "django-insecure-9yz$rw8%)1wm-l)j6q-r&$bu_n52sv=4q6)c5u8n10+5w+anec"  # noqa: S105

CLASS_C_NETWORK_ADDR = ["192.168.1.188"]
CLASS_C_DEVICE_ADDRS = [*CLASS_C_NETWORK_ADDR, "192.168.1.152", "192.168.1.240"]

INTERNAL_IPS = [*getenv("INTERNAL_IPS", "").split(), "localhost", "127.0.0.1"]
with contextlib.suppress(Exception):
    public_ip = get_public_ip()
    if public_ip:
        INTERNAL_IPS.append(str(public_ip))

INTERNAL_IPS += CLASS_C_DEVICE_ADDRS
ALLOWED_HOSTS += CLASS_C_NETWORK_ADDR

if getenv("DJANGO_MANAGEMENT_COMMAND", "").startswith("runserver"):
    logger.warning(
        "Loading picata.settings.dev…\n"
        f"INTERNAL_IPS = {INTERNAL_IPS}\n"
        f"ALLOWED_HOSTS = {ALLOWED_HOSTS}"
    )

    RUNSERVERPLUS_POLLER_RELOADER_TYPE = "watchdog"
    RUNSERVER_PLUS_EXCLUDE_PATTERNS = [
        ".venv/*",
        ".vscode/*",
        "build/*",
        "infra/*",
        "node_modules/*",
        "lib/*",
        "logs/*",
        "media/*",
        "snapshots/*",
        "src/migrations/*",
        "src/static/*",
        "static/*",
    ]

    # Enable Django Debug Toolbar and runserver_plus
    INSTALLED_APPS += [
        "debug_toolbar",
        "django_extensions",
    ]
    MIDDLEWARE = ["debug_toolbar.middleware.DebugToolbarMiddleware", *MIDDLEWARE]


# Create staticfiles.json manifest and hashed files when collecting static files
if getenv("DJANGO_MANAGEMENT_COMMAND") == "collectstatic":
    STORAGES["staticfiles"]["BACKEND"] = (
        "django.contrib.staticfiles.storage.ManifestStaticFilesStorage"
    )


# Enable extra information in template contexts for debugging
TEMPLATES[0]["OPTIONS"]["debug"] = True


# Logging (tuned for debugging)
LOGGING["handlers"]["console"]["level"] = "DEBUG"

LOGGING["root"]["level"] = "DEBUG"

LOGGING["loggers"]["gunicorn"] = {
    "handlers": ["console"],
    "level": "INFO",
    "propagate": False,
}

LOGGING["loggers"]["picata"]["level"] = "DEBUG"

LOGGING["loggers"]["django"]["level"] = "DEBUG"
LOGGING["loggers"]["django.db.backends"] = {"level": "INFO"}
LOGGING["loggers"]["django.template"]["level"] = "INFO"
LOGGING["loggers"]["django.request"]["level"] = "DEBUG"
LOGGING["loggers"]["django.request"]["handlers"] += ["console"]
LOGGING["loggers"]["django.utils"] = {"level": "INFO"}

LOGGING["loggers"]["wagtail"]["level"] = "DEBUG"

LOGGING["loggers"]["asyncio"] = {"level": "INFO"}



"""Django settings for test environments."""

from .base import *  # noqa: F403

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-9yz$rw8%)1wm-l)j6q-r&$bu_n52sv=4q6)c5u8n10+5w+anec"  # noqa: S105



"""Top-level URL configuration for the site."""

from debug_toolbar.toolbar import debug_toolbar_urls
from django.conf import settings
from django.contrib import admin
from django.urls import include, path, re_path
from wagtail import urls as wagtail_urls
from wagtail.admin import urls as wagtailadmin_urls
from wagtail.contrib.sitemaps.views import sitemap
from wagtail.documents import urls as wagtaildocs_urls
from wagtail.images.views.serve import ServeView

from picata.views import AtomArticleFeed, RSSArticleFeed, search

urlpatterns = [
    path("django-admin/", admin.site.urls),  # Django Admin
    path("admin/", include(wagtailadmin_urls)),  # Wagtail Admin
    path("documents/", include(wagtaildocs_urls)),  # Wagtail documents
    re_path(
        r"^images/([^/]*)/(\d*)/([^/]*)/[^/]*$", ServeView.as_view(), name="wagtailimages_serve"
    ),
    path("sitemap.xml", sitemap),
    path("feeds/rss/", RSSArticleFeed(), name="rss_feed"),
    path("feeds/atom/", AtomArticleFeed(), name="atom_feed"),
    path("search/", search, name="search"),
]

# Debug-mode-only URLs
if settings.DEBUG:
    from django.conf.urls.static import static
    from django.contrib.staticfiles.urls import staticfiles_urlpatterns
    from django.views.generic import RedirectView

    from picata.views import debug_shell, preview

    # Serve static and media files from development server
    urlpatterns += staticfiles_urlpatterns()
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

    # Enable Django Debug Toolbar
    urlpatterns += debug_toolbar_urls()

    urlpatterns += [
        path("favicon.ico", RedirectView.as_view(url=settings.STATIC_URL + "favicon.ico")),
        path("shell/", debug_shell),  # Just raises an exception (to invoke Werkzeug shell access)
        path("preview/<slug:file>/", preview, name="debug_preview"),  # templates/previews/<file>
    ]

# Let Wagtail take care of the rest
urlpatterns += [path("", include(wagtail_urls))]



"""Filter to transform Python types into human-readable strings."""  # noqa: INP001


def stringify(value: list, quote_style: str | None = None) -> str:
    """Convert a list of strings into a human-readable string with optional quoting."""
    if not value:
        return ""

    if not (isinstance(value, list | set | tuple)):
        raise TypeError("The 'stringify' filter currently only supports lists.")

    quote = "'" if quote_style == "single" else '"' if quote_style == "double" else ""
    quoted_items = [f"{quote}{item!s}{quote}" for item in value]

    if len(quoted_items) == 0:
        return ""
    if len(quoted_items) == 1:
        return quoted_items[0]
    if len(quoted_items) == 2:  # noqa: PLR2004
        return " and ".join(quoted_items)
    return f"{', '.join(quoted_items[:-1])}, and {quoted_items[-1]}"



"""Django settings for production environments."""
# ruff: noqa: F405

from .base import *  # noqa: F403

# ManifestStaticFilesStorage is recommended in production, to prevent
# outdated JavaScript / CSS assets being served from cache
# (e.g. after a Wagtail upgrade).
# See https://docs.djangoproject.com/en/5.1/ref/contrib/staticfiles/#manifeststaticfilesstorage
STORAGES["staticfiles"] = {
    "BACKEND": "django.contrib.staticfiles.storage.ManifestStaticFilesStorage"
}



"""Wagtail hooks, used to customise view-level behaviour of the Wagtail admin and front-end.

See: https://docs.wagtail.org/en/stable/reference/hooks.html
"""

import logging
from typing import ClassVar

from django.db.models import QuerySet, Value
from django.db.models.functions import Coalesce
from django.http import HttpRequest
from wagtail import hooks
from wagtail.models import Page
from wagtail.snippets.views.snippets import SnippetViewSet

from picata.models import PageTag

logger = logging.getLogger(__name__)


@hooks.register("construct_explorer_page_queryset")  # type: ignore[reportOptionalCall]
def order_admin_menu_by_date(parent_page: Page, pages: QuerySet, request: HttpRequest) -> QuerySet:  # noqa: ARG001
    """Order admin menus latest-at-top for pages with lots of children (like 'blog')."""
    if parent_page.slug == "blog":
        # Sort directly in order_by with Coalesce
        return pages.order_by(
            Coalesce("first_published_at", "latest_revision_created_at", Value("1970-01-01")).desc()
        )
    return pages


class PageTagViewSet(SnippetViewSet):
    """Viewset for managing `PageTag`s."""

    icon: str = "tag"
    list_display: ClassVar[list[str]] = ["name"]
    search_fields: ClassVar[list[str]] = ["name"]


@hooks.register("register_admin_viewset")  # type: ignore[reportOptionalCall]
def register_article_tag_viewset() -> SnippetViewSet:
    """Make `PageTag`s editable via the Wagtail admin."""
    return PageTagViewSet(model=PageTag)



"""Package for custom Django template tags."""

from django import template

from .filters.kill_orphans import killorphans
from .filters.stringify import stringify
from .tags.absolute_static import absolute_static
from .tags.menu_tags import render_site_menu

register = template.Library()

register.filter()(killorphans)
register.filter()(stringify)

register.simple_tag(takes_context=True)(absolute_static)
register.inclusion_tag(filename="picata/tags/site_menu.html", takes_context=True)(render_site_menu)

__all__ = ["absolute_static", "killorphans", "render_site_menu", "stringify"]



"""Simple template tag to produce absolute URLs from static file requests."""  # noqa: INP001

from django.templatetags.static import static

from picata.typing import Context


def absolute_static(context: Context, file: str) -> str:
    """Return the absolute path to a static file."""
    request = context["request"]
    return request.build_absolute_uri(static(file))



"""Top-level views for the site."""

# NB: Django's meta-class shenanigans over-complicate type hinting when QuerySets get involved.
# pyright: reportAttributeAccessIssue=false, reportArgumentType=false

import logging
from datetime import datetime
from typing import TYPE_CHECKING, NoReturn

from django.contrib.syndication.views import Feed
from django.http import HttpRequest, HttpResponse
from django.shortcuts import render
from django.utils.feedgenerator import Atom1Feed, Rss201rev2Feed

from picata.helpers.wagtail import (
    filter_pages_by_tags,
    filter_pages_by_type,
    page_preview_data,
    visible_pages_qs,
)
from picata.models import Article, ArticleType

if TYPE_CHECKING:
    from wagtail.query import PageQuerySet

logger = logging.getLogger(__name__)


class PostsFeed(Feed):
    """Base class for RSS and Atom article feeds."""

    title = "hpk.io Articles"
    link = "https://hpk.io/blog/"
    description = "Latest posts on hpk.io"

    def items(self) -> list[Article]:
        """Return the latest 10 published articles."""
        return list(Article.objects.live().order_by("-first_published_at"))

    def item_title(self, item: Article) -> str:
        """Return the article title."""
        return item.title

    def item_link(self, item: Article) -> str:
        """Return the absolute URL for the article."""
        return item.full_url

    def item_description(self, item: Article) -> str:
        """Return the article body as HTML with absolute URLs."""
        return item.content

    def item_pubdate(self, item: Article) -> datetime:
        """Return the article creation date."""
        return item.first_published_at


class RSSArticleFeed(PostsFeed):
    """RSS feed for articles."""

    feed_type = Rss201rev2Feed


class AtomArticleFeed(PostsFeed):
    """Atom feed for articles."""

    feed_type = Atom1Feed


def debug_shell(request: HttpRequest) -> NoReturn:
    """Just `assert False`, to force an exception and get to the Werkzeug debug console."""
    logger.info(
        "Raising `assert False` in the `debug_shell` view. "
        "Request details: method=%s, path=%s, user=%s",
        request.method,
        request.path,
        request.user if request.user.is_authenticated else "Anonymous",
    )
    assert False  # noqa: B011, PT015, S101


def preview(request: HttpRequest, file: str) -> HttpResponse:
    """Render a named template from the "templates/previews/" directory."""
    return render(request, f"picata/previews/{file}.html")


def search(request: HttpRequest) -> HttpResponse:
    """Render search results from the `query` and `tags` GET parameters."""
    results: dict[str, str | list[str] | set[str]] = {}

    # Base QuerySet for all pages
    pages: PageQuerySet = visible_pages_qs(request)

    # Perform search by query
    query_string = request.GET.get("query")
    if query_string:
        pages = pages.search(query_string)
        results["query"] = query_string

    # Resolve specific pages post-search
    specific_pages = [page.specific for page in pages]

    # Filter by page types
    page_types_string = request.GET.get("page_types")
    if page_types_string:
        page_type_slugs = {slug.strip() for slug in page_types_string.split(",") if slug.strip()}
        matching_page_types = ArticleType.objects.filter(slug__in=page_type_slugs)
        specific_pages = filter_pages_by_type(specific_pages, page_type_slugs)
        results["page_types"] = [page_type.name for page_type in matching_page_types]

    # Filter by tags
    tags_string = request.GET.get("tags")
    if tags_string:
        tags = {tag.strip() for tag in tags_string.split(",") if tag.strip()}
        specific_pages = filter_pages_by_tags(specific_pages, tags)
        results["tags"] = tags

    # Handle empty cases
    if not (query_string or tags_string or page_types_string):
        specific_pages = []

    # Enhance pages with preview and publication data
    page_previews = [page_preview_data(page, request) for page in specific_pages]

    return render(request, "picata/search_results.html", {**results, "pages": page_previews})



"""Django settings for the project.

Generated by 'django-admin startproject' using Django 5.1.

For more information on this file, see
https://docs.djangoproject.com/en/5.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/5.1/ref/settings/
"""

import contextlib
from os import getenv
from pathlib import Path

from picata.helpers import get_public_ip
from picata.log_utils import FormatterWithEverything

SRC_DIR = Path(__file__).resolve().parent.parent.parent
BASE_DIR = Path(SRC_DIR).parent
LOG_DIR = BASE_DIR / "logs"

INTERNAL_IPS = ["*"]

ALLOWED_HOSTS = [*getenv("FQDN", "").split(), "localhost", "127.0.0.1"]
with contextlib.suppress(Exception):
    public_ip = get_public_ip()
    if public_ip:
        ALLOWED_HOSTS.append(str(public_ip))

SECRET_KEY = getenv("SECRET_KEY")


# Application definition

INSTALLED_APPS = [
    # Local apps
    "picata.apps.Config",
    # Wagtail
    "wagtail.contrib.forms",
    "wagtail.contrib.redirects",
    "wagtail.contrib.routable_page",
    "wagtail.contrib.settings",
    "wagtail.embeds",
    "wagtail.sites",
    "wagtail.users",
    "wagtail.snippets",
    "wagtail.documents",
    "wagtail.images",
    "wagtail.search",
    "wagtail.admin",
    "wagtail_modeladmin",
    "wagtail",
    "modelcluster",
    "taggit",
    # Django apps
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sitemaps",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "wagtail.contrib.redirects.middleware.RedirectMiddleware",
    "picata.middleware.HTMLProcessingMiddleware",
]

ROOT_URLCONF = "picata.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [SRC_DIR / "templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "wagtail.contrib.settings.context_processors.settings",
            ],
        },
    },
]

WSGI_APPLICATION = "picata.wsgi.application"


# Database
# https://docs.djangoproject.com/en/5.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": getenv("PICATA_DB"),
        "USER": getenv("DB_USER"),
        "PASSWORD": getenv("DB_PASSWORD"),
        "HOST": getenv("DB_HOST", "localhost"),
        "PORT": "5432",
    },
}


# Password validation
# https://docs.djangoproject.com/en/5.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Logging
# https://docs.djangoproject.com/en/5.1/topics/logging/

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "complete": {
            "()": FormatterWithEverything,
        },
        "verbose": {
            "format": "----\n{levelname} {asctime} {name}:{module}\n{message}",
            "style": "{",
        },
        "simple": {
            "format": "{levelname} {module}: {message}",
            "style": "{",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "verbose",
            "level": "INFO",
        },
        "django_log": {
            "level": "INFO",
            "class": "picata.log_utils.RotatingDailyFileHandler",
            "filename": LOG_DIR / "django.log",
            "formatter": "verbose",
        },
        "picata_log": {
            "level": "INFO",
            "class": "picata.log_utils.RotatingDailyFileHandler",
            "filename": LOG_DIR / "picata.log",
            "formatter": "verbose",
        },
        "warnings_log": {
            "level": "WARNING",
            "class": "logging.handlers.RotatingFileHandler",
            "filename": LOG_DIR / "warnings.log",
            "maxBytes": 1024 * 1024 * 5,
            "backupCount": 5,
            "formatter": "verbose",
        },
    },
    "loggers": {
        "picata": {
            "handlers": ["picata_log"],
            "level": "INFO",
            "propagate": True,
        },
        "django": {
            "handlers": ["django_log", "warnings_log"],
            "level": "INFO",
            "propagate": True,
        },
        "django.request": {
            "handlers": ["django_log", "warnings_log"],
            "level": "WARNING",
            "propagate": False,
        },
        "django.template": {
            "handlers": ["django_log", "warnings_log"],
            "level": "WARNING",
            "propagate": False,
        },
        "wagtail": {
            "handlers": ["django_log", "warnings_log"],
            "level": "WARNING",
            "propagate": True,
        },
        "gunicorn.error": {
            "handlers": ["console"],
            "level": "ERROR",
            "propagate": False,
        },
        "watchdog.observers.inotify_buffer": {
            "handlers": ["console"],
            "level": "WARNING",
            "propagate": False,
        },
    },
    "root": {
        "handlers": ["console"],
        "level": "INFO",
    },
}


# Internationalization
# See https://docs.djangoproject.com/en/5.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = getenv("TIMEZONE", "UTC")

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/5.1/howto/static-files/

STATICFILES_FINDERS = [
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
]

STATICFILES_DIRS = [
    SRC_DIR / "static",
    BASE_DIR / "build/webpack",
]

STATIC_ROOT = BASE_DIR / "static"
STATIC_URL = "/static/"

MEDIA_ROOT = BASE_DIR / "media"
MEDIA_URL = "/media/"

# Default storage settings, with the staticfiles storage updated.
# See https://docs.djangoproject.com/en/5.1/ref/settings/#std-setting-STORAGES
STORAGES = {
    "default": {
        "BACKEND": "django.core.files.storage.FileSystemStorage",
    },
    "staticfiles": {
        "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
    },
}


# Email
# See https://docs.djangoproject.com/en/5.1/topics/email/

EMAIL_BACKEND = "django.core.mail.backends.smtp.EmailBackend"
EMAIL_HOST = "smtp.gmail.com"
EMAIL_PORT = 587
EMAIL_USE_TLS = True
EMAIL_HOST_USER = getenv("ADMIN_EMAIL")
EMAIL_HOST_PASSWORD = getenv("GMAIL_PASSWORD")
DEFAULT_FROM_EMAIL = getenv("ADMIN_EMAIL")


# Wagtail
# See https://docs.wagtail.org/en/stable/

WAGTAIL_SITE_NAME = "[Example Site Name]"

# Image serving
WAGTAILIMAGES_IMAGE_MODEL = "wagtailimages.Image"
WAGTAILIMAGES_SERVE_METHOD = "wagtail.images.views.serve.ServeView.as_view"

# Search - https://docs.wagtail.org/en/stable/topics/search/backends.html
WAGTAILSEARCH_BACKENDS = {
    "default": {
        "BACKEND": "wagtail.search.backends.database",
    },
}

# Base URL to use when referring to full URLs within the Wagtail admin backend -
# e.g. in notification emails. Don't include '/admin' or a trailing slash
WAGTAILADMIN_BASE_URL = "https://" + getenv("FQDN", "example.com")

# https://docs.wagtail.org/en/stable/reference/settings.html#general-editing
WAGTAILADMIN_RICH_TEXT_EDITORS = {
    "default": {
        "WIDGET": "wagtail.admin.rich_text.DraftailRichTextArea",
        "OPTIONS": {
            "features": [
                "h1",
                "h2",
                "h3",
                "h4",
                "h5",
                "h6",
                "bold",
                "italic",
                "ol",
                "ul",
                "code",
                "blockquote",
                "hr",
                "link",
                "document-link",
                "image",
                "embed",
                "superscript",
                "subscript",
                "strikethrough",
            ]
        },
    },
}

# Allowed file extensions for documents in the document library.
# This can be omitted to allow all files, but note that this may present a security risk
# if untrusted users are allowed to upload files -
# see https://docs.wagtail.org/en/stable/advanced_topics/deploying.html#user-uploaded-files
WAGTAILDOCS_EXTENSIONS = [
    "csv",
    "docx",
    "key",
    "odt",
    "pdf",
    "pptx",
    "rtf",
    "txt",
    "xlsx",
    "zip",
]



"""Generic helper-functions."""
# NB: Django's meta-class shenanigans over-complicate type hinting when QuerySets get involved.
# pyright: reportAttributeAccessIssue=false

from typing import cast

from django.http import HttpRequest
from wagtail.models import Page
from wagtail.query import PageQuerySet

from picata.models import TaggedPage

from . import get_models_of_type

TAGGED_PAGE_TYPES = get_models_of_type(TaggedPage)


def visible_pages_qs(request: HttpRequest) -> PageQuerySet:
    """Return a QuerySet of all pages derived from `Page` visible to the user."""
    pages = cast(PageQuerySet, Page.objects.all())
    if not request.user.is_authenticated:
        pages = pages.live()
    return pages


def filter_pages_by_tags(pages: list[Page], tags: set[str]) -> list[TaggedPage]:
    """Filter a list of pages to those containing all of a list of tags."""
    filtered_pages = []
    for page in pages:
        try:
            if isinstance(page, tuple(TAGGED_PAGE_TYPES)):
                page_tags = {tag.name for tag in page.tags.all()}
                if set(tags).issubset(page_tags):
                    filtered_pages.append(page)
        except AttributeError:
            continue
    return filtered_pages


def filter_pages_by_type(pages: list[Page], page_type_slugs: set[str]) -> list[Page]:
    """Filter a list of pages to those with a `page_type` matching any of the given slugs."""
    filtered_pages = []
    for page in pages:
        try:
            if (
                hasattr(page, "page_type")
                and page.page_type
                and page.page_type.slug in page_type_slugs
            ):
                filtered_pages.append(page)
        except AttributeError:
            continue
    return filtered_pages


def page_preview_data(page: Page, request: HttpRequest | None) -> dict[str, str]:
    """Return a dictionary of available publication and preview data for a page."""
    page_data = getattr(page, "preview_data", {}).copy()
    if hasattr(page, "get_publication_data"):
        page_data.update(page.get_publication_data(request))
    return page_data



"""Callables to transform the response."""

from lxml import etree

from picata.helpers import ALPHANUMERIC_REGEX, get_full_text


def add_heading_ids(tree: etree._Element) -> None:
    """Add a unique id to any heading in <main> missing one, derived from its inner text."""
    seen_ids = set()
    main = tree.xpath("/html/body/main")
    if not main:
        return

    for heading in main[0].xpath(".//h1|//h2|//h3|//h4|//h5|//h6"):
        if heading.get("id"):
            continue
        heading_text = get_full_text(heading)
        slug = heading_text.lower().replace(" ", "-")
        unique_id = slug
        count = 1
        while unique_id in seen_ids:
            unique_id = f"{slug}-{count}"
            count += 1
        seen_ids.add(unique_id)
        heading.set("id", unique_id)


class AnchorInserter:
    """Transformer to insert anchored pilcrows into targeted elements in the document."""

    def __init__(self, root: str, targets: str) -> None:
        """Remember the root paths we're to operate on."""
        self.root_xpath = root
        self.targets_xpath = targets

    def __call__(self, tree: etree._Element) -> None:
        """Inserts anchors into targets within the specified roots."""
        for root_element in tree.xpath(self.root_xpath):
            self._process_targets(root_element, self.targets_xpath)

    def _process_targets(self, root: etree._Element, targets: str) -> None:
        """Processes targets within a given root element, inserting anchors."""
        for target in root.xpath(targets):
            target_id = target.get("id")
            if not target_id or target.xpath(".//a"):
                continue

            sanitized_id = self._sanitize_id(target_id)
            if sanitized_id != target_id:
                target.set("id", sanitized_id)

            # Append an anchored pilcrow to the target element
            anchor = etree.Element("a", href=f"#{target_id}", **{"class": "target-link"})
            anchor.text = "¶"
            target.append(anchor)

    def _sanitize_id(self, id_value: str) -> str:
        """Sanitize the ID by removing non-alphanumeric characters."""
        return ALPHANUMERIC_REGEX.sub("", id_value)



"""HTML-processing middlware; should be placed last in (at the heart of) MIDDLEWARE."""

import logging
import re
from collections.abc import Callable
from typing import ClassVar

from django.http import HttpRequest, HttpResponse
from lxml import etree

from picata.helpers import make_response

logger = logging.getLogger(__name__)


class HTMLProcessingMiddleware:
    """Middleware register for text/html document transformers."""

    transformers: ClassVar[list[Callable[[etree._Element], None]]] = []

    def __init__(self, get_response: Callable) -> None:
        """Standard middleware initialisation; get the get_response method."""
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        """Filter the response through every registered transformer function."""
        response = self.get_response(request)

        if "text/html" in response.get("Content-Type", ""):
            doctype = self.extract_doctype(response.content.decode())
            tree = etree.fromstring(response.content, etree.HTMLParser())  # noqa: S320
            if tree is not None:
                for transformer in self.transformers:
                    transformer(tree)
                processed_html = etree.tostring(
                    tree,
                    pretty_print=True,  # type: ignore [reportCallIssue]
                    method="html",  # type: ignore [reportCallIssue]
                    encoding=str,  # type: ignore [reportCallIssue]
                )
                return make_response(response, f"{doctype}\n{processed_html}")

        return response

    @staticmethod
    def extract_doctype(html: str) -> str:
        """Extract the DOCTYPE declaration from the HTML."""
        match = re.match(r"^(<!DOCTYPE [^>]+>)", html, re.IGNORECASE)
        return match.group(1) if match else ""

    @classmethod
    def add_transformer(cls, func: Callable[[etree._Element], None]) -> None:
        """Add a transformation function to the global pipeline."""
        cls.transformers.append(func)



"""Django models; mostly subclassed Wagtail classes."""

# NB: Django's meta-class shenanigans over-complicate type hinting when QuerySets get involved.
# pyright: reportAttributeAccessIssue=false

from collections import OrderedDict
from datetime import timedelta
from typing import Any, ClassVar, cast

from django.contrib.auth.models import AnonymousUser, User
from django.db.models import (
    CASCADE,
    SET_NULL,
    CharField,
    ForeignKey,
    Model,
    SlugField,
    TextField,
)
from django.db.models.functions import Coalesce, ExtractYear
from django.http import HttpRequest
from django.urls import reverse
from modelcluster.contrib.taggit import ClusterTaggableManager
from modelcluster.fields import ParentalKey
from taggit.models import TagBase, TaggedItemBase
from wagtail.admin.panels import FieldPanel, Panel
from wagtail.blocks import RichTextBlock
from wagtail.contrib.routable_page.models import RoutablePageMixin
from wagtail.contrib.settings.models import BaseSiteSetting, register_setting
from wagtail.fields import RichTextField, StreamField
from wagtail.images.blocks import ImageChooserBlock
from wagtail.images.models import Image
from wagtail.models import Page, PageManager
from wagtail.query import PageQuerySet
from wagtail.search import index
from wagtail_modeladmin.options import ModelAdmin

from picata.typing import Args, Kwargs
from picata.typing.wagtail import PageContext

from .blocks import (
    CodeBlock,
    StaticIconLinkListsBlock,
    WrappedImageChooserBlock,
)


class BasePageContext(PageContext, total=False):
    """Return-type for an `Article`'s context dictionary."""

    url: str
    published: bool | str
    updated: bool | str
    latest_draft: str
    draft_url: str
    title: str


class BasePage(Page):
    """Mixin for `Page`-types offering previews of themselves on other `Page`s."""

    @property
    def preview_data(self) -> dict[str, Any]:
        """Return a dictionary of data used in previewing this page type."""
        return {
            "title": self.seo_title or self.title,
            "summary": f"<p>{self.search_description}</p>",
        }

    def get_publication_data(self, request: HttpRequest | None = None) -> dict[str, str]:
        """Helper method to calculate and format relevant dates for previews."""
        site = self.get_site()
        last_edited = self.latest_revision.created_at
        year = self.first_published_at.year if self.first_published_at else last_edited.year
        published, updated = self.first_published_at, self.last_published_at

        # Convert datetime objects to strings like "3 Jan, '25", or False, and
        # give a grace-period of one week for edits before marking the post as "updated"
        published_str = f"{published.day} {published:%b '%y}" if published else False
        updated_str = (
            f"{updated.day} {updated:%b '%y}"
            if published and updated and (updated >= published + timedelta(weeks=1))
            else False
        )

        data = {
            "year": year,
            "url": self.relative_url(site),
            "published": published_str,
            "updated": updated_str,
        }

        # Add last draft date & preview URL if there's an unpublished draft, for logged-in users
        if (
            (request and request.user.is_authenticated)
            and (not published or (updated and last_edited > updated))
            and hasattr(self, "id")
        ):
            data.update(
                {
                    "latest_draft": f"{last_edited.day} {last_edited:%b '%y}",
                    "draft_url": reverse("wagtailadmin_pages:preview_on_edit", args=[self.id]),
                }
            )

        return data

    def get_context(self, request: HttpRequest, *args: Args, **kwargs: Kwargs) -> BasePageContext:
        """Gather any publication and preview data available for the page into the context."""
        from picata.helpers.wagtail import page_preview_data

        context = super().get_context(request, *args, **kwargs)
        context.update(page_preview_data(self, request))
        return cast(BasePageContext, {**context})

    class Meta:
        """Declare `BasePage` as an abstract `Page` class."""

        abstract = True


# @register_snippet
class PageTag(TagBase):
    """Custom tag model for articles."""

    def __str__(self) -> str:
        """String representation of the tag."""
        return self.name


class PageTagRelation(TaggedItemBase):
    """Associates an PageTag with an Page."""

    tag: ForeignKey[PageTag] = ForeignKey(
        PageTag,
        related_name="tagged_items",
        on_delete=CASCADE,
    )
    content_object = ParentalKey(
        "Article",
        on_delete=CASCADE,
        related_name="tagged_items",
    )


class TaggedPage(BasePage):
    """Abstract base for a `Page` type supporting tags."""

    tags = ClusterTaggableManager(
        through=PageTagRelation,
        blank=True,
        help_text="Tags for the article.",
    )

    content_panels: ClassVar[list[Panel]] = [
        *BasePage.content_panels,
        FieldPanel("tags"),
    ]

    class Meta:
        """Declare `BasePage` as an abstract `Page` class."""

        abstract = True


class BasicPage(BasePage):
    """A basic page model for static content."""

    template = "picata/basic_page.html"

    content = StreamField(
        [
            ("rich_text", RichTextBlock()),
            ("code", CodeBlock()),
            ("image", ImageChooserBlock()),
        ],
        use_json_field=True,
        blank=True,
        help_text="Main content for the page.",
    )

    content_panels: ClassVar[list[FieldPanel]] = [
        *BasePage.content_panels,
        FieldPanel("content"),
    ]

    search_fields: ClassVar[list[index.SearchField]] = [
        *Page.search_fields,
        index.SearchField("content"),
    ]


class SplitViewPage(BasePage):
    """A page with 50%-width divs, split down the middle."""

    template = "picata/split_view.html"

    content = StreamField(
        [
            ("rich_text", RichTextBlock()),
            ("code", CodeBlock()),
            ("image", WrappedImageChooserBlock()),
            ("icon_link_lists", StaticIconLinkListsBlock()),
        ],
        use_json_field=True,
        blank=True,
        help_text="Main content for the split-view page.",
    )

    content_panels: ClassVar[list[FieldPanel]] = [
        *BasePage.content_panels,
        FieldPanel("content"),
    ]

    search_fields: ClassVar[list[index.SearchField]] = [
        *Page.search_fields,
        index.SearchField("content"),
    ]

    class Meta:
        """Declare explicit human-readable names for the page type."""

        verbose_name = "split-view page"
        verbose_name_plural = "split-view pages"


class ArticleType(Model):  # type: ignore[django-manager-missing]
    """Defines a type of article, like Blog Post, Review, or Guide."""

    name = CharField(max_length=100, unique=True, help_text="Name of the article type.")
    _Pluralised_name = CharField(
        max_length=100,
        blank=True,
        help_text="Plural form of the article type name (optional). Defaults to appending 's'.",
    )
    slug = SlugField(unique=True, max_length=100)
    description = TextField(blank=True, help_text="Optional description of this type.")

    def __str__(self) -> str:
        """Return the name of the ArticleType."""
        return self.name

    @property
    def name_plural(self) -> str:
        """Return the plural name of the article type."""
        return self._Pluralised_name or f"{self.name}s"

    @property
    def indefinite_article(self) -> str:
        """Return a string like 'a guide' or 'an article'."""
        name_lower = self.name.lower()
        return f"{'an' if name_lower[0] in 'aeiou' else 'a'} {name_lower}"


class ArticleTypeAdmin(ModelAdmin):
    """Wagtail admin integration for managing article types."""

    model = ArticleType
    menu_label = "Article Types"  # Label for the menu item
    menu_icon = "tag"  # Icon for the menu item (from Wagtail icon set)
    add_to_settings_menu = True  # Whether to add to the "Settings" menu
    list_display = ("name", "slug")  # Fields to display in the listing
    search_fields = ("name", "slug")  # Fields to include in the search bar


class ArticleContext(BasePageContext):
    """Return-type for an `Article`'s context dictionary."""

    content: str


class ArticleQuerySet(PageQuerySet):
    """Default `QuerySet` for all `Article`-type pages."""

    def with_effective_date(self) -> PageQuerySet:
        """Annotate with 'effective_date' to allow date-ordering to consider recent drafts."""
        return self.annotate(
            effective_date=Coalesce("first_published_at", "latest_revision_created_at")
        )

    def by_date(self) -> PageQuerySet:
        """Return all `Article` pages, ordered by decending "effective" date."""
        return self.with_effective_date().order_by("-effective_date")

    def live_for_user(self, user: AnonymousUser | User) -> PageQuerySet:
        """Filter out non-live pages for non-authenticated users."""
        return self if user.is_authenticated else self.live()


class Article(TaggedPage):
    """Class for article-like pages."""

    template = "picata/article.html"
    objects = PageManager.from_queryset(ArticleQuerySet)()

    tagline: CharField = CharField(blank=True, help_text="A short tagline for the article.")
    summary = RichTextField(blank=True, help_text="A summary to be displayed in previews.")
    content = StreamField(
        [
            ("rich_text", RichTextBlock()),
            ("code", CodeBlock()),
            ("image", ImageChooserBlock()),
        ],
        use_json_field=True,
        blank=True,
        help_text="Main content for the article.",
    )

    page_type: ForeignKey[ArticleType | None] = ForeignKey(
        ArticleType,
        null=True,
        blank=True,
        on_delete=SET_NULL,
        related_name="articles",
        help_text="Select the type of article.",
    )

    content_panels: ClassVar[list[Panel]] = [
        *TaggedPage.content_panels,
        FieldPanel("tagline"),
        FieldPanel("summary"),
        FieldPanel("content"),
        FieldPanel("page_type"),
    ]

    search_fields: ClassVar[list[index.SearchField]] = [
        *TaggedPage.search_fields,
        index.SearchField("tagline"),
        index.SearchField("summary"),
        index.SearchField("content"),
        index.SearchField("tags"),
        index.SearchField("page_type"),
    ]

    @property
    def preview_data(self) -> dict[str, Any]:
        """Return data required to render a preview of this article."""
        return {
            **super().preview_data,
            "tagline": self.tagline,
            "summary": self.summary,
            "page_type": self.page_type,
            "tags": list(self.tags.all()),
        }

    def get_context(self, request: HttpRequest, *args: Args, **kwargs: Kwargs) -> ArticleContext:
        """Provide extra context needed for the `Article` to render itself."""
        context = dict(super().get_context(request, *args, **kwargs))
        context.update({"content": self.content})
        return cast(ArticleContext, context)


class PostGroupePageContext(PageContext):
    """Return-type for a `PostGroupPage`'s context dictionary."""

    posts: OrderedDict[int, list[dict[str, str]]]


class PostGroupPage(RoutablePageMixin, Page):
    """A top-level page for grouping various types of posts or articles."""

    template = "picata/post_listing.html"
    subpage_types: ClassVar[list[str]] = ["picata.Article"]

    intro = RichTextField(blank=True, help_text="An optional introduction to this group.")

    content_panels: ClassVar[list[Panel]] = [*BasePage.content_panels, FieldPanel("intro")]

    def get_context(
        self, request: HttpRequest, *args: Args, **kwargs: Kwargs
    ) -> PostGroupePageContext:
        """Add a dictionary of posts grouped by year to the context dict."""
        children = self.get_children()
        if not request.user.is_authenticated:
            children = children.live()
        children = children.specific()
        children = children.annotate(
            effective_date=Coalesce("first_published_at", "latest_revision_created_at"),
            year_published=ExtractYear("first_published_at"),
        )

        # Create an OrderedDict grouping posts by year in reverse chronological order
        posts_by_year: OrderedDict = OrderedDict()
        for child in children.order_by("-effective_date"):
            post_data = getattr(child, "preview_data", {}).copy()
            post_data.update(**child.get_publication_data(request))

            # Group posts by year, defaulting to last-draft year if unpublished
            if post_data["year"] not in posts_by_year:
                posts_by_year[post_data["year"]] = []
            posts_by_year[post_data["year"]].append(post_data)

        return cast(
            PostGroupePageContext,
            {**super().get_context(request, *args, **kwargs), "posts_by_year": posts_by_year},
        )

    class Meta:
        """Declare more human-friendly names for the page type."""

        verbose_name: str = "post listing"
        verbose_name_plural: str = "post listings"


@register_setting
class SocialSettings(BaseSiteSetting):
    """Site-wide social media configuration."""

    default_social_image: ForeignKey[Image] = ForeignKey(
        Image,
        null=True,
        blank=True,
        on_delete=SET_NULL,
        help_text="Default image for social media previews.",
        related_name="+",
    )

    panels: ClassVar[list[Panel]] = [
        FieldPanel("default_social_image"),
    ]


class HomePageContext(BasePageContext):
    """Return-type for the `HomePage`'s context dictionary."""

    top_content: str
    bottom_content: str
    recent_posts: list[BasePage]


class HomePage(BasePage):
    """Single-use specialised page for the root of the site."""

    template = "picata/home_page.html"

    top_content = StreamField(
        [
            ("rich_text", RichTextBlock()),
            ("image", WrappedImageChooserBlock()),
            ("icon_link_lists", StaticIconLinkListsBlock()),
        ],
        use_json_field=True,
        blank=True,
        help_text="Content stream above 'Recent posts'",
    )

    bottom_content = StreamField(
        [
            ("rich_text", RichTextBlock()),
            ("image", WrappedImageChooserBlock()),
            ("icon_link_lists", StaticIconLinkListsBlock()),
        ],
        use_json_field=True,
        blank=True,
        help_text="Content stream rendered under 'Recent posts'",
    )

    content_panels: ClassVar[list[FieldPanel]] = [
        *BasePage.content_panels,
        FieldPanel("top_content"),
        FieldPanel("bottom_content"),
    ]

    search_fields: ClassVar[list[index.SearchField]] = [
        *Page.search_fields,
        index.SearchField("top_content"),
        index.SearchField("bottom_content"),
    ]

    def get_context(self, request: HttpRequest, *args: Args, **kwargs: Kwargs) -> HomePageContext:
        """Add content streams and a recent posts list to the context."""
        from picata.helpers.wagtail import page_preview_data

        recent_posts = Article.objects.live_for_user(request.user).by_date()
        recent_posts = [page_preview_data(post, request) for post in recent_posts]

        return cast(
            HomePageContext,
            {
                **dict(super().get_context(request, *args, **kwargs)),
                "top_content": self.top_content,
                "bottom_content": self.bottom_content,
                "recent_posts": recent_posts,
            },
        )

    class Meta:
        """Declare explicit human-readable names for the page type."""

        verbose_name = "home page"



# Generated by Django 5.1.5 on 2025-01-24 00:18

import django.db.models.deletion
import modelcluster.contrib.taggit
import modelcluster.fields
import wagtail.contrib.routable_page.models
import wagtail.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0094_alter_page_locale'),
        ('wagtailimages', '0027_image_description'),
    ]

    operations = [
        migrations.CreateModel(
            name='ArticleType',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Name of the article type.', max_length=100, unique=True)),
                ('_Pluralised_name', models.CharField(blank=True, help_text="Plural form of the article type name (optional). Defaults to appending 's'.", max_length=100)),
                ('slug', models.SlugField(max_length=100, unique=True)),
                ('description', models.TextField(blank=True, help_text='Optional description of this type.')),
            ],
        ),
        migrations.CreateModel(
            name='BasicPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.page')),
                ('content', wagtail.fields.StreamField([('rich_text', 0), ('code', 3), ('image', 4)], blank=True, block_lookup={0: ('wagtail.blocks.RichTextBlock', (), {}), 1: ('wagtail.blocks.TextBlock', (), {'help_text': None, 'required': True}), 2: ('wagtail.blocks.ChoiceBlock', [], {'choices': [('python', 'Python'), ('javascript', 'JavaScript'), ('html', 'HTML'), ('css', 'CSS'), ('bash', 'Bash'), ('plaintext', 'Plain Text')], 'required': False}), 3: ('wagtail.blocks.StructBlock', [[('code', 1), ('language', 2)]], {}), 4: ('wagtail.images.blocks.ImageChooserBlock', (), {})}, help_text='Main content for the page.')),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='HomePage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.page')),
                ('top_content', wagtail.fields.StreamField([('rich_text', 0), ('image', 1), ('icon_link_lists', 11)], blank=True, block_lookup={0: ('wagtail.blocks.RichTextBlock', (), {}), 1: ('picata.blocks.WrappedImageChooserBlock', (), {}), 2: ('wagtail.blocks.CharBlock', (), {'help_text': 'Optional heading for this list (e.g., Social Links).', 'required': False}), 3: ('wagtail.blocks.IntegerBlock', (), {'default': 2, 'help_text': 'Heading level for the list (1-6).', 'max_value': 6, 'min_value': 1, 'required': False}), 4: ('picata.blocks.HREFBlock', (), {'help_text': 'An optional link field', 'max_length': 255, 'required': False}), 5: ('wagtail.blocks.CharBlock', (), {'help_text': 'The title for the list item.', 'max_length': 50, 'required': True}), 6: ('wagtail.blocks.CharBlock', (), {'help_text': 'Id of the icon in the static/icons.svg file.', 'max_length': 255, 'required': False}), 7: ('wagtail.blocks.StructBlock', [[('href', 4), ('label', 5), ('icon', 6)]], {}), 8: ('wagtail.blocks.ListBlock', (7,), {'help_text': 'The list of items.'}), 9: ('wagtail.blocks.StructBlock', [[('heading', 2), ('heading_level', 3), ('items', 8)]], {}), 10: ('wagtail.blocks.StreamBlock', [[('link_list', 9)]], {'help_text': 'Add one or more heading-and-link-list blocks.', 'required': False}), 11: ('wagtail.blocks.StructBlock', [[('lists', 10)]], {})}, help_text="Content stream above 'Recent posts'")),
                ('bottom_content', wagtail.fields.StreamField([('rich_text', 0), ('image', 1), ('icon_link_lists', 11)], blank=True, block_lookup={0: ('wagtail.blocks.RichTextBlock', (), {}), 1: ('picata.blocks.WrappedImageChooserBlock', (), {}), 2: ('wagtail.blocks.CharBlock', (), {'help_text': 'Optional heading for this list (e.g., Social Links).', 'required': False}), 3: ('wagtail.blocks.IntegerBlock', (), {'default': 2, 'help_text': 'Heading level for the list (1-6).', 'max_value': 6, 'min_value': 1, 'required': False}), 4: ('picata.blocks.HREFBlock', (), {'help_text': 'An optional link field', 'max_length': 255, 'required': False}), 5: ('wagtail.blocks.CharBlock', (), {'help_text': 'The title for the list item.', 'max_length': 50, 'required': True}), 6: ('wagtail.blocks.CharBlock', (), {'help_text': 'Id of the icon in the static/icons.svg file.', 'max_length': 255, 'required': False}), 7: ('wagtail.blocks.StructBlock', [[('href', 4), ('label', 5), ('icon', 6)]], {}), 8: ('wagtail.blocks.ListBlock', (7,), {'help_text': 'The list of items.'}), 9: ('wagtail.blocks.StructBlock', [[('heading', 2), ('heading_level', 3), ('items', 8)]], {}), 10: ('wagtail.blocks.StreamBlock', [[('link_list', 9)]], {'help_text': 'Add one or more heading-and-link-list blocks.', 'required': False}), 11: ('wagtail.blocks.StructBlock', [[('lists', 10)]], {})}, help_text="Content stream rendered under 'Recent posts'")),
            ],
            options={
                'verbose_name': 'home page',
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='PageTag',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True, verbose_name='name')),
                ('slug', models.SlugField(allow_unicode=True, max_length=100, unique=True, verbose_name='slug')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PostGroupPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.page')),
                ('intro', wagtail.fields.RichTextField(blank=True, help_text='An optional introduction to this group.')),
            ],
            options={
                'verbose_name': 'post listing',
                'verbose_name_plural': 'post listings',
            },
            bases=(wagtail.contrib.routable_page.models.RoutablePageMixin, 'wagtailcore.page'),
        ),
        migrations.CreateModel(
            name='SplitViewPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.page')),
                ('content', wagtail.fields.StreamField([('rich_text', 0), ('code', 3), ('image', 4), ('icon_link_lists', 14)], blank=True, block_lookup={0: ('wagtail.blocks.RichTextBlock', (), {}), 1: ('wagtail.blocks.TextBlock', (), {'help_text': None, 'required': True}), 2: ('wagtail.blocks.ChoiceBlock', [], {'choices': [('python', 'Python'), ('javascript', 'JavaScript'), ('html', 'HTML'), ('css', 'CSS'), ('bash', 'Bash'), ('plaintext', 'Plain Text')], 'required': False}), 3: ('wagtail.blocks.StructBlock', [[('code', 1), ('language', 2)]], {}), 4: ('picata.blocks.WrappedImageChooserBlock', (), {}), 5: ('wagtail.blocks.CharBlock', (), {'help_text': 'Optional heading for this list (e.g., Social Links).', 'required': False}), 6: ('wagtail.blocks.IntegerBlock', (), {'default': 2, 'help_text': 'Heading level for the list (1-6).', 'max_value': 6, 'min_value': 1, 'required': False}), 7: ('picata.blocks.HREFBlock', (), {'help_text': 'An optional link field', 'max_length': 255, 'required': False}), 8: ('wagtail.blocks.CharBlock', (), {'help_text': 'The title for the list item.', 'max_length': 50, 'required': True}), 9: ('wagtail.blocks.CharBlock', (), {'help_text': 'Id of the icon in the static/icons.svg file.', 'max_length': 255, 'required': False}), 10: ('wagtail.blocks.StructBlock', [[('href', 7), ('label', 8), ('icon', 9)]], {}), 11: ('wagtail.blocks.ListBlock', (10,), {'help_text': 'The list of items.'}), 12: ('wagtail.blocks.StructBlock', [[('heading', 5), ('heading_level', 6), ('items', 11)]], {}), 13: ('wagtail.blocks.StreamBlock', [[('link_list', 12)]], {'help_text': 'Add one or more heading-and-link-list blocks.', 'required': False}), 14: ('wagtail.blocks.StructBlock', [[('lists', 13)]], {})}, help_text='Main content for the split-view page.')),
            ],
            options={
                'verbose_name': 'split-view page',
                'verbose_name_plural': 'split-view pages',
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='Article',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.page')),
                ('tagline', models.CharField(blank=True, help_text='A short tagline for the article.')),
                ('summary', wagtail.fields.RichTextField(blank=True, help_text='A summary to be displayed in previews.')),
                ('content', wagtail.fields.StreamField([('rich_text', 0), ('code', 3), ('image', 4)], blank=True, block_lookup={0: ('wagtail.blocks.RichTextBlock', (), {}), 1: ('wagtail.blocks.TextBlock', (), {'help_text': None, 'required': True}), 2: ('wagtail.blocks.ChoiceBlock', [], {'choices': [('python', 'Python'), ('javascript', 'JavaScript'), ('html', 'HTML'), ('css', 'CSS'), ('bash', 'Bash'), ('plaintext', 'Plain Text')], 'required': False}), 3: ('wagtail.blocks.StructBlock', [[('code', 1), ('language', 2)]], {}), 4: ('wagtail.images.blocks.ImageChooserBlock', (), {})}, help_text='Main content for the article.')),
                ('page_type', models.ForeignKey(blank=True, help_text='Select the type of article.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='articles', to='picata.articletype')),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='PageTagRelation',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('content_object', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='tagged_items', to='picata.article')),
                ('tag', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tagged_items', to='picata.pagetag')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='article',
            name='tags',
            field=modelcluster.contrib.taggit.ClusterTaggableManager(blank=True, help_text='Tags for the article.', through='picata.PageTagRelation', to='picata.PageTag', verbose_name='Tags'),
        ),
        migrations.CreateModel(
            name='SocialSettings',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('default_social_image', models.ForeignKey(blank=True, help_text='Default image for social media previews.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailimages.image')),
                ('site', models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE, to='wagtailcore.site')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
